#include "Mesh.hpp"
#include "Viewer.hpp"
#include <cstdlib>
#include <unordered_set>
#include <stack>

int
usage(int argc, char * argv[])
{
  DGP_CONSOLE << "";
  DGP_CONSOLE << "Usage: " << argv[0] << " <mesh>";
  DGP_CONSOLE << "";

  return -1;
}


int const NUM_PALETTE_COLORS = 24;
ColorRGB COLOR_PALETTE[NUM_PALETTE_COLORS] = {
  ColorRGB::fromARGB(0x298edb),
  ColorRGB::fromARGB(0x982411),
  ColorRGB::fromARGB(0x6d4e25),
  ColorRGB::fromARGB(0x1b5043),
  ColorRGB::fromARGB(0x6e7662),
  ColorRGB::fromARGB(0xa08b00),
  ColorRGB::fromARGB(0x58427b),
  ColorRGB::fromARGB(0x1d2f5b),
  ColorRGB::fromARGB(0xac5e34),
  ColorRGB::fromARGB(0x804055),
  ColorRGB::fromARGB(0x6d7a00),
  ColorRGB::fromARGB(0x572e2c),

  // Invert each color above
  ColorRGB::fromARGB(~0x298edb),
  ColorRGB::fromARGB(~0x982411),
  ColorRGB::fromARGB(~0x6d4e25),
  ColorRGB::fromARGB(~0x1b5043),
  ColorRGB::fromARGB(~0x6e7662),
  ColorRGB::fromARGB(~0xa08b00),
  ColorRGB::fromARGB(~0x58427b),
  ColorRGB::fromARGB(~0x1d2f5b),
  ColorRGB::fromARGB(~0xac5e34),
  ColorRGB::fromARGB(~0x804055),
  ColorRGB::fromARGB(~0x6d7a00),
  ColorRGB::fromARGB(~0x572e2c)
};

int
numPaletteColors()
{
  return NUM_PALETTE_COLORS;
}

ColorRGB const &
getPaletteColor(long i)
{
  long n = numPaletteColors();
  long index;
  if (i < 0)
    index = (n - ((-i) % n)) % n;
  else
    index = i % n;

  return COLOR_PALETTE[index];
}

long
colorMeshComponents(Mesh & mesh)
{
  typedef std::unordered_set<Mesh::Face *> FaceSet;
  typedef std::stack<Mesh::Face *> FaceStack;

  FaceSet unvisited;
  for (Mesh::FaceIterator fi = mesh.facesBegin(); fi != mesh.facesEnd(); ++fi)
    unvisited.insert(&(*fi));

  long n = 0;
  while (!unvisited.empty())
  {
    FaceStack stack;

    Mesh::Face * face = *unvisited.begin();
    stack.push(face);
    unvisited.erase(face);

    ColorRGBA component_color = getPaletteColor(n);

    // Depth-first search
    while (!stack.empty())
    {
      face = stack.top();
      stack.pop();

      face->setColor(component_color);

      for (Mesh::Face::EdgeIterator fei = face->edgesBegin(); fei != face->edgesEnd(); ++fei)
      {
        Mesh::Edge * edge = *fei;
        for (Mesh::Edge::FaceIterator efi = edge->facesBegin(); efi != edge->facesEnd(); ++efi)
        {
          Mesh::Face * nbr = *efi;
          if (unvisited.find(nbr) != unvisited.end())
          {
            stack.push(nbr);
            unvisited.erase(nbr);
          }
        }
      }
    }

    n++;
  }

  return n;
}

int
main(int argc, char * argv[])
{
  if (argc < 2)
    return usage(argc, argv);

  std::string in_path = argv[1];

  Mesh mesh;
  if (!mesh.load(in_path))
    return -1;

  long num_components = colorMeshComponents(mesh);

  DGP_CONSOLE << "Read " << num_components << "-component mesh '" << mesh.getName() << "' with " << mesh.numVertices()
              << " vertices, " << mesh.numEdges() << " edges and " << mesh.numFaces() << " faces from " << in_path;

  Viewer viewer;
  viewer.setObject(&mesh);
  viewer.launch(argc, argv);

  return 0;
}
